#include <stdio.h>
#include <stdlib.h>
#include <pthread.h>
#include <malloc.h>
#include <libspe2.h>

#include <tnw_ppu.h>
#include "commands.h"

extern spe_program_handle_t spu_prg; 

typedef struct {

    spe_context_ptr_t ctx;
    pthread_t pthread;
    pthread_attr_t attr;
    unsigned int entry;
    unsigned int flags;
    void *argp;
    void *envp;
    spe_stop_info_t stopinfo;
    
} spe_thread_info;

static inline void mail_blk(spe_context_ptr_t ctx, unsigned int msg) {

  spe_in_mbox_write(ctx, &msg, 1, SPE_MBOX_ANY_BLOCKING);


}

static void *ppu_thread(void *arg) {
    
    spe_thread_info *datap = (spe_thread_info *)arg;
    int rc;
    
    do {
        rc = spe_context_run(datap->ctx, &datap->entry, datap->flags,
        datap->argp, datap->envp, &datap->stopinfo);
        
    }while (rc > 0); /* until exit or error, while stop & signal */
    
    pthread_exit(NULL);
}


spe_thread_info *spe_thread_start(spe_program_handle_t spu_program, 
			    unsigned int flags, void *argp, void *envp) {

  spe_gang_context_ptr_t gang=NULL;
  spe_thread_info *thread_data;

  flags = flags | SPE_MAP_PS;
  
  thread_data = malloc(sizeof(spe_thread_info));
  thread_data->flags = flags;
  thread_data->entry = SPE_DEFAULT_ENTRY;
  thread_data->ctx = spe_context_create(flags, gang);
  spe_program_load(thread_data->ctx, &spu_program);
  
  thread_data->argp = argp;
  thread_data->envp = envp;
  
  pthread_create(&thread_data->pthread, NULL, &ppu_thread, thread_data);

  return thread_data;
}


void spe_thread_join(spe_thread_info *thread_info) {

  pthread_join(thread_info->pthread, NULL);
  spe_context_destroy(thread_info->ctx);
  free(thread_info);
  
}



tnw_rx_t rx;





int main(int argc, char *argv[]) {

  spe_thread_info *thread_info[8];
  int nthreads=8;
  int i;
  uint cmd;  
  uint use_link[6];
  uint nlink=(argc-1);
  spe_context_ptr_t ctx;
  

  
  printf("nlink = %d\n", nlink);
  for(i=0;i<nlink;i++) {
	  use_link[i]=atoi(argv[i+1]);
  }

  /* initialization of tnw *BEFORE* STARTING SPU THREADS */
  tnw_init();


  /* get tnw informations */

  tnw_nc_pos_t abs_pos;
  

  abs_pos = tnw_get_nc_pos();
  printf("absolute nc position bp = %d nc = %d\n", abs_pos.bp, abs_pos.nc);

  /*get topology info*/
  
  tnw_topo_t topo;
  char *recv_buf;
  unsigned int rbraddr;

  tnw_get_topology(&topo);
  
  printf("red sz %d green sz %d blu sz %d\n", topo.r_size, topo.g_size, topo.b_size);
  printf("red co %d green co %d blu co %d\n", topo.r_coord, topo.g_coord, topo.b_coord);
  
  printf("rp %d rm %d gp %d gm %d bp %d bm %d\n", topo.rp_lnk, topo.rm_lnk, topo.gp_lnk, topo.gm_lnk,topo.bp_lnk, topo.bm_lnk);

  for(i=0;i<nthreads;i++) {
    thread_info[i] = spe_thread_start(spu_prg, SPE_CFG_SIGNOTIFY1_OR | SPE_CFG_SIGNOTIFY2_OR, 0, 0);
  }
  
  printf("started %d threads\n", nthreads);
  
  //allocate one page;
  recv_buf = memalign(65536, 65536);

  //get real address of the page and pin the page
  rbraddr = tnw_lock_page((unsigned int)recv_buf);


  //init rx handler;
  tnw_init_rx(&rx, use_link[0], 7);
  
  //configure credit base
  tnw_credit_base(use_link[0], 7, rbraddr);
  
  //give the credit for data, 128B
  tnw_credit(&rx, 0, 128, 0);
  
  cmd =  SPU_CMD_SEND;
  if(nlink > 1) nlink = 1;
  for(i=0;i<nlink;i++) { 
    ctx = thread_info[i]->ctx;
    mail_blk(ctx, cmd);
    mail_blk(ctx, use_link[i]); // link
    mail_blk(ctx, 7); // channel
  }



  //wait for notify
  tnw_wait_notify(&rx, 0);
  
  //check arrived data
  for(i=0;i<128;i++) {
    printf("%d ", recv_buf[i]);
    
  }
  printf("\n");

  //  unpin page
  tnw_unlock_page(rbraddr);
  
  
  cmd =  SPU_CMD_EXIT;  
  for(i=0;i<nthreads;i++) {
    mail_blk(thread_info[i]->ctx, cmd);
  }
  for(i=0;i<nthreads;i++) {
    spe_thread_join(thread_info[i]);
  }
  printf("threads joined ciaooo!!!\n");

  return 0;
}
