#include <stdio.h>
#include <stdlib.h>
#include <pthread.h>
#include <libspe2.h>

#include <tnw_ppu.h>
#include <torus_ppu.h>

#include "commands.h"

extern spe_program_handle_t spu_prg; 

typedef struct {

    spe_context_ptr_t ctx;
    pthread_t pthread;
    pthread_attr_t attr;
    unsigned int entry;
    unsigned int flags;
    void *argp;
    void *envp;
    spe_stop_info_t stopinfo;
    
} spe_thread_info;

static inline void mail_blk(spe_context_ptr_t ctx, unsigned int msg) {

  spe_in_mbox_write(ctx, &msg, 1, SPE_MBOX_ANY_BLOCKING);


}

static void *ppu_thread(void *arg) {
    
    spe_thread_info *datap = (spe_thread_info *)arg;
    int rc;
    
    do {
        rc = spe_context_run(datap->ctx, &datap->entry, datap->flags,
        datap->argp, datap->envp, &datap->stopinfo);
        
    }while (rc > 0); /* until exit or error, while stop & signal */
    
    pthread_exit(NULL);
}


spe_thread_info *spe_thread_start(spe_program_handle_t spu_program, 
			    unsigned int flags, void *argp, void *envp) {

  spe_gang_context_ptr_t gang=NULL;
  spe_thread_info *thread_data;

  flags = flags | SPE_MAP_PS;
  
  thread_data = malloc(sizeof(spe_thread_info));
  thread_data->flags = flags;
  thread_data->entry = SPE_DEFAULT_ENTRY;
  thread_data->ctx = spe_context_create(flags, gang);
  spe_program_load(thread_data->ctx, &spu_program);
  
  thread_data->argp = argp;
  thread_data->envp = envp;
  
  pthread_create(&thread_data->pthread, NULL, &ppu_thread, thread_data);

  return thread_data;
}


void spe_thread_join(spe_thread_info *thread_info) {

  pthread_join(thread_info->pthread, NULL);
  spe_context_destroy(thread_info->ctx);
  free(thread_info);
  
}


int main(int argc, char *argv[]) {

  spe_thread_info *thread_info[8];
  int nthreads=8;
  int i;
  uint cmd;  
  uint use_link[6];
  uint nlink=(argc-1);
  spe_context_ptr_t ctx;
  

  

  /* initialization of tnw *BEFORE* STARTING SPU THREADS */
  //         blue         red          green
  torus_init(TORUS_MAP_X, TORUS_MAP_Z, TORUS_MAP_Y);


  /* get tnw informations */

  tnw_nc_pos_t abs_pos;
  
  
  abs_pos = tnw_get_nc_pos();
  printf("absolute nc position bp = %d nc = %d\n", abs_pos.bp, abs_pos.nc);
  
  /* get topology info */
  int sx, sy, sz;
  torus_size(&sx, &sy, &sz);
  printf("sx=%d sy=%d sz=%d\n", sx, sy, sz);

  int cx, cy, cz;
  torus_coords(&cx, &cy, &cz);
  printf("cx=%d cy=%d cz=%d\n", cx, cy, cz);
  

  for(i=0;i<nthreads;i++) {
    thread_info[i] = spe_thread_start(spu_prg, SPE_CFG_SIGNOTIFY1_OR | SPE_CFG_SIGNOTIFY2_OR, 0, 0);
  }
  
  printf("started %d threads\n", nthreads);
 
  nlink = 1;

  cmd =  SPU_CMD_RECV;
  for(i=0;i<nlink;i++) { 
    ctx = thread_info[i]->ctx;
    mail_blk(ctx, cmd);
    mail_blk(ctx, 7); // channel
  }

  //sleep(100);
  
  cmd =  SPU_CMD_EXIT;  
  for(i=0;i<nthreads;i++) {
    mail_blk(thread_info[i]->ctx, cmd);
  }
  for(i=0;i<nthreads;i++) {
    spe_thread_join(thread_info[i]);
  }
  printf("threads joined ciaooo!!!\n");

  return 0;
}
