#ifndef TNW_PPU_H
#define TNW_PPU_H



#include "tnw.h"

typedef enum { TNW_IF_PRIMARY, TNW_IF_REDUNDANT } tnw_if_mode_t;

typedef struct {

  tnw_nc_pos_t apos0;
  tnw_nc_pos_t apos1;

  unsigned int bc;

  tnw_nc_pos_t apos;

  int r_size;
  int g_size;
  int b_size;

  int r_coord;
  int g_coord;
  int b_coord;


  int rp_lnk;
  int rm_lnk;

  tnw_if_mode_t rp_ifm;
  tnw_if_mode_t rm_ifm;
  
  int gp_lnk;
  int gm_lnk;

  tnw_if_mode_t gp_ifm;
  tnw_if_mode_t gm_ifm;

  int bp_lnk;
  int bm_lnk;

  tnw_if_mode_t bp_ifm;
  tnw_if_mode_t bm_ifm;

} tnw_topo_t;

typedef struct {
  volatile vector unsigned short buf;
  unsigned int phy;
  unsigned int ch;
  unsigned short state;
  unsigned int buf_raddr;
} tnw_rx_t  __attribute__((aligned(128)));

/*
typedef struct {

  int r_size;
  int g_size;
  int b_size;

  int r_coord;
  int g_coord;
  int b_coord;
  
  int rp_lnk;
  int rm_lnk;
  
  int gp_lnk;
  int gm_lnk;

  int bp_lnk;
  int bm_lnk;

} tnw_topo_t;
*/

/* initialization */
int tnw_init();



/* memory mapped low level access */
void tnw_reg_write(unsigned int addr, unsigned int data);
unsigned int tnw_reg_read(unsigned int addr);

void tnw_rx_write(unsigned int phy, unsigned int reg, unsigned int data);
unsigned int tnw_rx_read(unsigned int phy, unsigned int reg);

void tnw_tx_write(unsigned int phy, unsigned int reg, unsigned int data);
unsigned int tnw_tx_read(unsigned int phy, unsigned int reg);


unsigned int tnw_mdio_read(unsigned int phy, unsigned int reg);
void tnw_mdio_write(unsigned int phy, unsigned int reg, unsigned int data);

/* abs nc position */
tnw_nc_pos_t tnw_get_nc_pos();

/* user topology information  */
unsigned int tnw_get_topo(tnw_topo_t *topo);
void tnw_get_topology(tnw_topo_t *topo);


unsigned int tnw_lock_page(unsigned int addr);
unsigned int tnw_unlock_page(unsigned int raddr);
void tnw_credit_base(unsigned int phy, unsigned int channel, unsigned int addr);
void tnw_init_rx(tnw_rx_t *rx, unsigned int phy, unsigned int channel);
void tnw_credit(tnw_rx_t *rx, unsigned int addr, 
		unsigned int size, unsigned int ctag);
/* returns 1 if notify arrived, 0 if not arrived */
unsigned int tnw_test_notify(tnw_rx_t *rx, unsigned int ctag);
void tnw_wait_notify(tnw_rx_t *rx, unsigned int ctag);
tnw_rx_t *tnw_rx_alloc();
void tnw_rx_free(tnw_rx_t *rx);

#endif
